from formset.renderers.default import FormRenderer as DefaultFormRenderer


class FormRenderer(DefaultFormRenderer):
    max_options_per_line = 4

    def __init__(self, **kwargs):
        kwargs.setdefault('label_css_classes', 'label')
        super().__init__(**kwargs)

    _template_mapping = dict(DefaultFormRenderer._template_mapping, **{
        'django/forms/default.html': 'formset/bulma/form.html',
        'django/forms/widgets/checkbox.html': 'formset/bulma/widgets/checkbox.html',
        'django/forms/widgets/radio.html': 'formset/bulma/widgets/multiple_input.html',
        'django/forms/widgets/select.html': 'formset/bulma/widgets/select.html',
        'django/forms/widgets/checkbox_select.html': 'formset/bulma/widgets/multiple_input.html',
        'formset/default/widgets/selectize.html': 'formset/bulma/widgets/select.html',
        'formset/default/widgets/file.html': 'formset/bulma/widgets/file.html',
        'formset/default/widgets/dual_selector.html': 'formset/bulma/widgets/dual_selector.html',
    })

    def _amend_input(self, context):
        context['widget']['attrs']['class'] = 'input'
        return context

    def _amend_label(self, context):
        return super()._amend_label(context, hide_checkbox_label=True)

    def _amend_multiple_input(self, context, css_class):
        context = super()._amend_multiple_input(context)
        css_classes = [css_class]
        if not context['widget'].get('inlined_options'):
            css_classes.append('is-block ml-0 mb-1')
        for _, optgroup, _ in context['widget']['optgroups']:
            for option in optgroup:
                option['template_name'] = 'formset/bulma/widgets/input_option.html'
                option['label_css_classes'] = ' '.join(css_classes)
        return context

    def _amend_checkbox_select(self, context):
        return self._amend_multiple_input(context, 'checkbox mr-2')

    def _amend_radio(self, context):
        return self._amend_multiple_input(context, 'radio mr-1')

    def _amend_textarea(self, context):
        context['widget']['attrs']['class'] = 'textarea'
        return context

    def _amend_collection(self, context):
        context = super()._amend_collection(context)
        context.update({
            'add_collection_button': 'formset/bulma/buttons/add_collection.html',
            'remove_collection_button': 'formset/bulma/buttons/remove_collection.html',
        })
        return context

    _context_modifiers = dict(DefaultFormRenderer._context_modifiers, **{
        'django/forms/label.html': _amend_label,
        'django/forms/widgets/text.html': _amend_input,
        'django/forms/widgets/email.html': _amend_input,
        'django/forms/widgets/date.html': _amend_input,
        'django/forms/widgets/number.html': _amend_input,
        'django/forms/widgets/password.html': _amend_input,
        'django/forms/widgets/checkbox_select.html': _amend_checkbox_select,
        'django/forms/widgets/textarea.html': _amend_textarea,
        'django/forms/widgets/radio.html': _amend_radio,
        'formset/default/collection.html': _amend_collection,
    })



from django.core.exceptions import ImproperlyConfigured
from django.forms import forms

from formset.utils import FormMixin


class FieldsetMixin(FormMixin):
    show_if = None
    hide_if = None
    disable_if = None
    legend = None
    template_name = 'formset/default/fieldset.html'

    def __init__(self, **kwargs):
        show_if = kwargs.pop('show_if', None)
        hide_if = kwargs.pop('hide_if', None)
        disable_if = kwargs.pop('disable_if', None)
        if show_if and hide_if:
            raise ImproperlyConfigured(f"class {self.__class__} can accept either `show_if` or `hide_if`, but not both.")
        if show_if:
            self.show_if = show_if
        elif hide_if:
            self.hide_if = hide_if
        if disable_if:
            self.disable_if = disable_if
        legend = kwargs.pop('legend', None)
        if legend:
            self.legend = legend
        super().__init__(**kwargs)

    def get_context(self):
        context = super().get_context()
        context.update(
            show_if=self.show_if,
            hide_if=self.hide_if,
            disable_if=self.disable_if,
            legend=self.legend,
        )
        return context


class Fieldset(FieldsetMixin, forms.Form):
    """
    This is just DOM sugar wrapped into a Form. It therefore behaves like a Form object and should
    be used as such. Its purpose is to add visual elements to a `<form>`. Remember, a Form is just a
    data-abstraction layer, has no display properties and is not intended to be styled or anotated.
    On the other side, a <fieldset> may offer a `<legend>`, a border and the possibility to
    show/hide or disable a set of fields. A `HTMLFieldSetElement` however does not has any field
    validation functionality, this is left to the `HTMLFormElement`.
    """
    def __repr__(self):
        return f'<{self.__class__.__name__} legend="{self.legend}" template_name="{self.template_name}">'



from formset.renderers.default import FormRenderer as DefaultFormRenderer


class FormRenderer(DefaultFormRenderer):
    max_options_per_line = 4

    def __init__(self, **kwargs):
        kwargs.setdefault('label_css_classes', 'form-label')
        super().__init__(**kwargs)

    _template_mapping = dict(DefaultFormRenderer._template_mapping, **{
        'django/forms/default.html': 'formset/bootstrap/form.html',
        'django/forms/widgets/checkbox.html': 'formset/bootstrap/widgets/checkbox.html',
        'django/forms/widgets/radio.html': 'formset/bootstrap/widgets/multiple_input.html',
        'django/forms/widgets/checkbox_select.html': 'formset/bootstrap/widgets/multiple_input.html',
        'formset/default/widgets/dual_selector.html': 'formset/bootstrap/widgets/dual_selector.html',
        'formset/default/widgets/file.html': 'formset/bootstrap/widgets/file.html',
    })

    def _amend_input(self, context):
        context['widget']['attrs']['class'] = 'form-control'
        return context

    def _amend_checkbox(self, context):
        context['widget']['attrs']['class'] = 'form-check-input'
        return context

    def _amend_select(self, context):
        context['widget']['attrs']['class'] = 'form-select'
        return context

    def _amend_dual_selector(self, context):
        context.update(
            select_classes='form-select',
            lookup_field_classes='form-control form-control-sm',
        )
        return context

    def _amend_file(self, context):
        return context  # intentionally noop

    def _amend_label(self, context):
        return super()._amend_label(context, hide_checkbox_label=True)

    def _amend_multiple_input(self, context):
        context = super()._amend_multiple_input(context)
        for _, optgroup, _ in context['widget']['optgroups']:
            for option in optgroup:
                option['attrs']['class'] = 'form-check-input'
                option['template_name'] = 'formset/bootstrap/widgets/input_option.html'
        return context

    def _amend_collection(self, context):
        context = super()._amend_collection(context)
        context.update({
            'add_collection_button': 'formset/bootstrap/buttons/add_collection.html',
            'remove_collection_button': 'formset/bootstrap/buttons/remove_collection.html',
        })
        return context

    _context_modifiers = dict(DefaultFormRenderer._context_modifiers, **{
        'django/forms/label.html': _amend_label,
        'django/forms/widgets/text.html': _amend_input,
        'django/forms/widgets/email.html': _amend_input,
        'django/forms/widgets/date.html': _amend_input,
        'django/forms/widgets/number.html': _amend_input,
        'django/forms/widgets/password.html': _amend_input,
        'django/forms/widgets/textarea.html': _amend_input,
        'django/forms/widgets/select.html': _amend_select,
        'django/forms/widgets/checkbox.html': _amend_checkbox,
        'django/forms/widgets/checkbox_select.html': _amend_multiple_input,
        'django/forms/widgets/radio.html': _amend_multiple_input,
        'formset/default/widgets/file.html': _amend_file,
        'formset/default/widgets/selectize.html': _amend_select,
        'formset/default/widgets/dual_selector.html': _amend_dual_selector,
        'formset/default/collection.html': _amend_collection,
    })



from django import template
from django.forms import BaseForm
from django.middleware.csrf import get_token
from django.template.exceptions import TemplateSyntaxError
from django.utils.module_loading import import_string

from formset.utils import FormMixin, FormsetErrorList
from formset.renderers.default import FormRenderer


def _formsetify(form, *args, **kwargs):
    assert isinstance(form, BaseForm), \
        "Must be applied to a Form object inheriting from 'django.forms.BaseForm'."
    if not isinstance(form, FormMixin):
        form.__class__ = type(form.__class__.__name__, (FormMixin, form.__class__), {})

    renderer_args = [
        ('form_css_classes', kwargs.pop('form_classes', None)),
        ('fieldset_css_classes', kwargs.pop('fieldset_classes', None)),
        ('field_css_classes', kwargs.pop('field_classes', None)),
        ('label_css_classes', kwargs.pop('label_classes', None)),
        ('control_css_classes', kwargs.pop('control_classes', None)),
        ('form_css_classes', kwargs.pop('form_classes', None)),
        ('collection_css_classes', kwargs.pop('collection_classes', None)),
        ('max_options_per_line', kwargs.pop('max_options_per_line', None)),
    ]
    if len(kwargs):
        raise TemplateSyntaxError(f"Unknown argument '{kwargs.popitem()[0]}' in formsetify.")
    renderer_kwargs = {key: value for key, value in renderer_args if value is not None}

    if len(args) == 1 and args[0]:
        framework = args[0].lower()
        if '.' in framework:
            form.renderer = import_string(f'{framework}.FormRenderer')(**renderer_kwargs)
        else:
            form.renderer = import_string(f'formset.renderers.{framework}.FormRenderer')(**renderer_kwargs)
    elif not isinstance(form.renderer, FormRenderer):
        form.renderer = FormRenderer(**renderer_kwargs)
    form.error_class = FormsetErrorList
    return form


def formsetify(context, form, *args, **kwargs):
    get_token(context['request'])  # ensures that the CSRF-Cookie is set
    _formsetify(form, *args, **kwargs)
    return ''


def render_form(context, form, *args, **kwargs):
    get_token(context['request'])  # ensures that the CSRF-Cookie is set
    form = _formsetify(form, *args, **kwargs)
    return str(form)


register = template.Library()
register.simple_tag(formsetify, name='formsetify', takes_context=True)
register.simple_tag(render_form, name='render_form', takes_context=True)



from django.core import checks
from django.db.models.fields.related import ManyToManyField

from formset.widgets import DualSortableSelector


class SortableMultipleChoiceMixin:
    def clean(self, value):
        qs = super().clean(value)
        return qs, value


class SortableManyToManyField(ManyToManyField):
    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_ordering_enabled(**kwargs),
            *self._check_related_model(**kwargs),
        ]

    def _check_ordering_enabled(self, **kwargs):
        if not self.remote_field.through._meta.ordering:
            return [
                checks.Error(
                    "SortableManyToManyField must point to a mapping model with enabled ordering.",
                    obj=self,
                    id="fields.E341",
                )
            ]
        return []

    def _check_related_model(self, **kwargs):
        try:
            fields = self.remote_field.through._meta.fields
            next(iter(f for f in fields if getattr(f, 'related_model', None) is self.related_model))
        except StopIteration:
            return [
                checks.Error(
                    f"{self.remote_field.through} implementing the many-to-many relation must have a ForeignKey on '{self.related_model}'.",
                    obj=self,
                    id="fields.E342",
                )
            ]
        return []


    def value_from_object(self, obj):
        if obj.pk is None:
            return []
        # find the field in the M2M relation pointing on our related model ('to' in constructor)
        through = self.remote_field.through
        field = next(iter(f for f in through._meta.fields if getattr(f, 'related_model', None) is self.related_model))
        return [getattr(item, field.name) for item in through._default_manager.select_related(field.name)]

    def save_form_data(self, instance, data):
        qs, orig_value = data
        super().save_form_data(instance, qs)
        # after assigning the M2M relation the ordering is updated
        mrm = getattr(instance, self.attname)
        order_field_name = mrm.through._meta.ordering[0]
        for order, target_pk in enumerate(orig_value, 1):
            filter_kwargs = {
                mrm.source_field.attname: instance.pk,
                mrm.target_field.attname: target_pk,
            }
            mrm.through._default_manager.filter(**filter_kwargs).update(**{order_field_name: order})

    def formfield(self, **kwargs):
        form_field = super().formfield(**kwargs)
        form_field.__class__ = type(
            form_field.__class__.__name__,
            (SortableMultipleChoiceMixin, form_field.__class__),
            {}
        )
        if not isinstance(form_field.widget, DualSortableSelector):
            form_field.widget = DualSortableSelector()
        return form_field



import copy

from django.conf import settings
from django.core.exceptions import ValidationError
from django.forms.utils import ErrorDict, ErrorList
from django.utils.functional import cached_property

from formset.boundfield import BoundField
from formset.renderers.default import FormRenderer

MARKED_FOR_REMOVAL = '_marked_for_removal_'


class FormsetErrorList(ErrorList):
    template_name = 'formset/default/field_errors.html'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if args and hasattr(args[0], 'client_messages'):
            self.client_messages = args[0].client_messages

    def copy(self):
        copy = super().copy()
        if hasattr(self, 'client_messages'):
            copy.client_messages = self.client_messages
        return copy

    def get_context(self):
        return {
            'errors': self,
            'client_messages': self.client_messages,
        }


class HolderMixin:
    ignore_marked_for_removal = getattr(settings, 'FORMSET_IGNORE_MARKED_FOR_REMOVAL', False)

    def __init__(self, **kwargs):
        self.marked_for_removal = False
        super().__init__(**kwargs)

    def replicate(self, data=None, initial=None, prefix=None, renderer=None, ignore_marked_for_removal=None):
        replica = copy.copy(self)
        if hasattr(self, 'declared_holders'):
            replica.declared_holders = {
                key: holder.replicate(
                    renderer=renderer,
                    ignore_marked_for_removal=ignore_marked_for_removal,
                ) for key, holder in self.declared_holders.items()
            }
        replica.data = data
        replica.is_bound = data is not None
        replica._errors = None
        try:
            delattr(replica, 'cleaned_data')
        except AttributeError:
            pass
        if hasattr(replica, 'files'):
            replica.files.clear()
        if initial:
            replica.initial = initial
        if prefix:
            replica.prefix = prefix
        if ignore_marked_for_removal is not None:
            replica.ignore_marked_for_removal = ignore_marked_for_removal
        if isinstance(replica.renderer, FormRenderer):
            return replica
        if self.default_renderer:
            if isinstance(self.default_renderer, type):
                replica.renderer = self.default_renderer()
            else:
                replica.renderer = self.default_renderer
        elif renderer:
            replica.renderer = renderer
        else:
            replica.renderer = FormRenderer()
        return replica

    def _clean_for_removal(self):
        """
        Forms which have been marked for removal, clean their received form data,
        but always keep them as validated.
        """
        self._errors = ErrorDict()
        self.cleaned_data = {}
        for name, bf in self._bound_items():
            field = bf.field
            value = bf.initial if field.disabled else bf.data
            try:
                value = field.clean(value)
                if hasattr(self, f'clean_{name}'):
                    self.cleaned_data[name] = value
                    value = getattr(self, f'clean_{name}')()
            except ValidationError:
                pass  # ignore all validation errors for removed forms
            finally:
                self.cleaned_data[name] = value
        self.cleaned_data[MARKED_FOR_REMOVAL] = True
        self.marked_for_removal = True

    def is_valid(self):
        if self.is_bound and MARKED_FOR_REMOVAL in self.data:
            self._clean_for_removal()
            return True
        return super().is_valid()


class FormMixin(HolderMixin):
    """
    Mixin class to be added to a native Django Form. This is required to add
    """

    def __init__(self, error_class=FormsetErrorList, **kwargs):
        kwargs['error_class'] = error_class
        super().__init__(**kwargs)

    def __getitem__(self, name):
        "Returns a modified BoundField for the given field."
        try:
            field = self.fields[name]
        except KeyError:
            raise KeyError(f"Key {name} not found in Form")
        return BoundField(self, field, name)

    @cached_property
    def form_id(self):
        # The "form" tag is used to link fields to their form owner
        # See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input#attr-form for details
        if self.prefix and self.auto_id and '%s' in str(self.auto_id):
            return self.auto_id % self.prefix

    def add_prefix(self, field_name):
        """
        Return the field name with a prefix preended, if this Form has a prefix set.
        """
        return f'{self.prefix}.{field_name}' if self.prefix else field_name

    def get_context(self):
        """
        This simplified method just returns the ``form``, but not the ``fields``, ``hidden_fields``
        and ``errors``, since they are rendered by the included ``form.html`` template.
        """
        return {
            'form': self,
        }

    def get_field(self, field_name):
        return self.declared_fields[field_name]



from formset.renderers.default import FormRenderer as DefaultFormRenderer


class FormRenderer(DefaultFormRenderer):
    max_options_per_line = 4

    _template_mapping = dict(DefaultFormRenderer._template_mapping, **{
        'django/forms/default.html': 'formset/foundation/form.html',
        'django/forms/widgets/checkbox.html': 'formset/foundation/widgets/checkbox.html',
        'django/forms/widgets/radio.html': 'formset/foundation/widgets/multiple_input.html',
        'formset/default/widgets/file.html': 'formset/foundation/widgets/file.html',
        'formset/default/widgets/dual_selector.html': 'formset/foundation/widgets/dual_selector.html',
        'django/forms/widgets/checkbox_select.html': 'formset/foundation/widgets/multiple_input.html',
    })

    def _amend_label(self, context):
        return super()._amend_label(context, hide_checkbox_label=True)

    def _amend_multiple_input(self, context):
        context = super()._amend_multiple_input(context)
        if context['widget'].get('inlined_options'):
            for _, optgroup, _ in context['widget']['optgroups']:
                for option in optgroup:
                    option['template_name'] = 'formset/foundation/widgets/inlined_input_option.html'
        return context

    def _amend_collection(self, context):
        context = super()._amend_collection(context)
        context.update({
            'add_collection_button': 'formset/foundation/buttons/add_collection.html',
            'remove_collection_button': 'formset/foundation/buttons/remove_collection.html',
        })
        return context

    _context_modifiers = dict(DefaultFormRenderer._context_modifiers, **{
        'django/forms/label.html': _amend_label,
        'django/forms/widgets/checkbox_select.html': _amend_multiple_input,
        'django/forms/widgets/radio.html': _amend_multiple_input,
        'formset/default/collection.html': _amend_collection,
    })



import mimetypes
import os
from pathlib import Path
import tempfile

from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.files.storage import default_storage
from django.core.signing import get_cookie_signer
from django.http.response import HttpResponseBadRequest, JsonResponse


THUMBNAIL_MAX_HEIGHT = 200
THUMBNAIL_MAX_WIDTH = 350
UPLOAD_TEMP_DIR = Path(default_storage.base_location) / 'upload_temp'


def get_thumbnail_path(image_path, image_height=THUMBNAIL_MAX_HEIGHT):
    image_path = Path(image_path)
    thumbnail_name = f'{image_path.stem}_h{image_height}{image_path.suffix}'
    return image_path.with_name(thumbnail_name)


def thumbnail_image(storage, image_path, image_height=THUMBNAIL_MAX_HEIGHT):
    try:
        from PIL import Image, ImageOps

        image = Image.open(image_path)
    except Exception:
        return staticfiles_storage.url('formset/icons/file-picture.svg')
    else:
        height = int(image_height)
        width = int(round(image.width * height / image.height))
        width, height = min(width, THUMBNAIL_MAX_WIDTH), min(height, THUMBNAIL_MAX_HEIGHT)
        thumb = ImageOps.fit(ImageOps.exif_transpose(image), (width, height))
        thumbnail_path = get_thumbnail_path(image_path, image_height)
        thumb.save(thumbnail_path)
        return storage.url(thumbnail_path.relative_to(storage.location))


def split_mime_type(content_type):
    try:
        return content_type.split('/')
    except (AttributeError, ValueError):
        return "application", "octet-stream"


def file_icon_url(mime_type, sub_type):
    if mime_type in ['audio', 'font', 'video']:
        return staticfiles_storage.url(f'formset/icons/file-{mime_type}.svg')
    if mime_type == 'application' and sub_type in ['zip', 'pdf']:
        return staticfiles_storage.url(f'formset/icons/file-{sub_type}.svg')
    return staticfiles_storage.url('formset/icons/file-unknown.svg')


def get_file_info(field_file):
    if not field_file:
        return None
    file_path = Path(field_file.path)
    content_type, _ = mimetypes.guess_type(file_path)
    mime_type, sub_type = split_mime_type(content_type)
    if mime_type == 'image':
        if sub_type == 'svg+xml':
            thumbnail_url = field_file.url
        else:
            thumbnail_path = get_thumbnail_path(file_path)
            if thumbnail_path.is_file():
                thumbnail_url = field_file.storage.url(thumbnail_path.relative_to(field_file.storage.location))
            else:
                thumbnail_url = thumbnail_image(field_file.storage, file_path)
    else:
        thumbnail_url = file_icon_url(mime_type, sub_type)
    name = '.'.join(file_path.name.split('.')[1:])
    if file_path.is_file():
        download_url = field_file.url
        file_size = depict_size(field_file.size)
    else:
        download_url = 'javascript:void(0);'
        thumbnail_url = staticfiles_storage.url('formset/icons/file-missing.svg')
        file_size = '–'
    return {
        'content_type': content_type,
        'name': name,
        'path': field_file.name,
        'download_url': download_url,
        'thumbnail_url': thumbnail_url,
        'size': file_size,
    }


class FileUploadMixin:
    """
    Add this mixin class to views classes using forms which accept file uploads through
    the provided widget :class:`formset.widgets.UploadedFileInput`.
    """
    filename_max_length = 250

    def post(self, request, **kwargs):
        if request.content_type == 'multipart/form-data' and 'temp_file' in request.FILES and 'image_height' in request.POST:
            return self._receive_uploaded_file(request.FILES['temp_file'], request.POST['image_height'])
        return super().post(request, **kwargs)

    def _receive_uploaded_file(self, file_obj, image_height=None):
        """
        Iterate over all uploaded files.
        """
        if not file_obj:
            return HttpResponseBadRequest(f"File upload failed for '{file_obj.name}'.")
        signer = get_cookie_signer(salt='formset')

        # copy uploaded file into temporary clipboard inside the default storage location
        UPLOAD_TEMP_DIR.mkdir(parents=True, exist_ok=True)
        file_path = Path(file_obj.name)
        fh, temp_path = tempfile.mkstemp(suffix=file_path.suffix, prefix=f'{file_path.stem}.', dir=UPLOAD_TEMP_DIR)
        for chunk in file_obj.chunks():
            os.write(fh, chunk)
        os.close(fh)
        assert default_storage.size(temp_path) == file_obj.size
        relative_path = Path(temp_path).relative_to(default_storage.location)
        download_url = default_storage.url(relative_path)

        # dict returned by the form on submission
        mime_type, sub_type = split_mime_type(file_obj.content_type)
        if mime_type == 'image':
            if sub_type == 'svg+xml':
                thumbnail_url = download_url
            else:
                thumbnail_url = thumbnail_image(default_storage, temp_path, image_height=image_height)
        else:
            thumbnail_url = file_icon_url(mime_type, sub_type)
        file_handle = {
            'upload_temp_name': signer.sign(relative_path),
            'content_type': f'{mime_type}/{sub_type}',
            'content_type_extra': file_obj.content_type_extra,
            'name': file_obj.name[:self.filename_max_length],
            'download_url': download_url,
            'thumbnail_url': thumbnail_url,
            'size': file_obj.size,
        }
        return JsonResponse(file_handle)


def depict_size(size):
    if size > 1048576:
        return '{:.1f}MB'.format(size / 1048576)
    if size > 10240:
        return '{:.0f}kB'.format(size / 1024)
    return str(size)



import json

from django.http.response import HttpResponseBadRequest, JsonResponse
from django.utils.functional import cached_property
from django.views.generic.base import ContextMixin, TemplateResponseMixin, View
from django.views.generic.edit import FormView as GenericFormView
from django.views.generic.detail import SingleObjectMixin

from formset.upload import FileUploadMixin
from formset.widgets import Selectize, DualSelector


class IncompleteSelectResponseMixin:
    """
    Add this mixin class to views classes using forms with incomplete fields. These fields
    usually are of type ChoiceField referring to a foreign model and using one of the widgets
    :class:`formset.widgets.Selectize`, :class:`formset.widgets.SelectizeMultiple` or
    :class:`formset.widgets.DualSelector`.
    """
    def get(self, request, **kwargs):
        if request.accepts('application/json') and 'field' in request.GET:
            if 'query' in request.GET or 'offset' in request.GET:
                return self._fetch_options(request)
        return super().get(request, **kwargs)

    def _fetch_options(self, request):
        field_path = request.GET['field']
        try:
            field = self.get_field(field_path)
        except KeyError:
            return HttpResponseBadRequest(f"No such field: {field_path}")
        assert isinstance(field.widget, (Selectize, DualSelector))
        try:
            offset = int(request.GET.get('offset'))
        except TypeError:
            offset = 0
        if query := request.GET.get('query'):
            data = {'query': query}
            queryset = field.widget.search(query)
            incomplete = None  # incomplete state unknown
        else:
            data = {}
            queryset = field.widget.choices.queryset
            incomplete = queryset.count() - offset > field.widget.max_prefetch_choices
        limited_qs = queryset[offset:offset + field.widget.max_prefetch_choices]
        to_field_name = field.to_field_name if field.to_field_name else 'pk'
        items = [{'id': getattr(item, to_field_name), 'label': str(item)} for item in limited_qs]
        data.update(
            count=len(items),
            total_count=field.widget.choices.queryset.count(),
            incomplete=incomplete,
            items=items,
        )
        return JsonResponse(data)


class FormsetResponseMixin:
    @cached_property
    def _request_body(self):
        if self.request.content_type == 'application/json':
            return json.loads(self.request.body)

    def get_extra_data(self):
        """
        When submitting a form, one can additionally add extra parameters via the button's ``submit()`` action.
        Use this method to access that extra data.
        """
        if self._request_body:
            return self._request_body.get('_extra')


class FormViewMixin(FormsetResponseMixin):
    def get_success_url(self):
        """
        In **django-formset**, the success_url may be None and set inside the templates.
        """
        return str(self.success_url) if self.success_url else None

    def form_valid(self, form):
        response = super().form_valid(form)
        assert response.status_code == 302
        return JsonResponse({'success_url': self.get_success_url()})

    def form_invalid(self, form):
        super().form_invalid(form)
        return JsonResponse(form.errors, status=422, safe=False)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        if self._request_body:
            kwargs['data'] = self._request_body.get('formset_data')
        return kwargs

    def get_field(self, path):
        field_name = path.split('.')[-1]
        return self.form_class.base_fields[field_name]


class FormView(IncompleteSelectResponseMixin, FileUploadMixin, FormViewMixin, GenericFormView):
    """
    FormView class used as controller for handling a single Django Form. The purpose of this View
    is to render the provided Form, when invoked as a standard GET-request using the provided Django
    Template.
    This View also acts as endpoint for POST-requests submitting files, as endpoint for GET-requests
    querying for autocomplete Select-Fields and as endpoint for Form submissions.

    It can be used directly inside the URL routing:

    .. code-block:: python

        from django.urls import path
        from formset.views import FormView

        ...
        urlpatterns = [
            ...
            path('my-uri', FormView.as_view(
                form_class=MyForm,
                template_name='my-form.html',
                success_url='/success',
            )),
            ...
        ]

    or by inheriting from ``FormView`` and overwriting the attributes in that class:

    .. code-block:: python

        from formset.views import FormView

        class MyFormEditView(FormView):
            form_class = MyForm
            template_name = 'my-form.html'
            success_url = '/success'
    """


class FormCollectionViewMixin(FormsetResponseMixin):
    collection_class = None
    success_url = None
    initial = {}

    def get(self, request, *args, **kwargs):
        """Handle GET requests: instantiate blank versions of the forms in the collection."""
        return self.render_to_response(self.get_context_data())

    def post(self, request, **kwargs):
        form_collection = self.get_form_collection()
        if form_collection.is_valid():
            return self.form_collection_valid(form_collection)
        else:
            return self.form_collection_invalid(form_collection)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['form_collection'] = self.get_form_collection()
        return context

    def get_field(self, path):
        return self.form_collection.get_field(path)

    def get_form_collection(self):
        collection_class = self.get_collection_class()
        kwargs = {
            'initial': self.get_initial(),
        }
        if self.request.method in ('POST', 'PUT') and self.request.content_type == 'application/json':
            body = json.loads(self.request.body)
            kwargs.update(data=body.get('formset_data'))
        return collection_class(**kwargs)

    def get_collection_class(self):
        return self.collection_class

    def get_initial(self):
        """Return the initial data to use for collections of forms on this view."""
        return self.initial.copy()

    def get_success_url(self):
        return str(self.success_url) if self.success_url else None

    def form_collection_valid(self, form_collection):
        return JsonResponse({'success_url': self.get_success_url()})

    def form_collection_invalid(self, form_collection):
        return JsonResponse(form_collection.errors, status=422, safe=False)


class FormCollectionView(IncompleteSelectResponseMixin, FileUploadMixin, FormCollectionViewMixin, ContextMixin, TemplateResponseMixin, View):
    pass


class EditCollectionView(IncompleteSelectResponseMixin, FileUploadMixin, FormCollectionViewMixin, SingleObjectMixin, TemplateResponseMixin, View):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_initial(self):
        initial = super().get_initial()
        if isinstance(initial, dict) and self.object:
            collection_class = self.get_collection_class()
            initial.update(collection_class().model_to_dict(self.object))
        return initial

    def form_collection_valid(self, form_collection):
        form_collection.construct_instance(self.object, form_collection.cleaned_data)
        return super().form_collection_valid(form_collection)



from django.core.exceptions import NON_FIELD_ERRORS
from django.forms.forms import BaseForm
from django.forms.models import BaseModelForm, construct_instance, model_to_dict
from django.forms.widgets import MediaDefiningClass
from django.forms.utils import ErrorDict, ErrorList, RenderableMixin
from django.utils.datastructures import MultiValueDict

from formset.exceptions import FormCollectionError
from formset.renderers.default import FormRenderer
from formset.utils import HolderMixin, FormMixin, FormsetErrorList, MARKED_FOR_REMOVAL

COLLECTION_ERRORS = '_collection_errors_'


class FormCollectionMeta(MediaDefiningClass):
    """
    Collect Forms declared on the base classes.
    """
    def __new__(cls, name, bases, attrs):
        # Collect forms and sub-collections from current class and remove them from attrs.
        attrs['declared_holders'] = {}
        for key, value in list(attrs.items()):
            if isinstance(value, (BaseForm, BaseFormCollection)):
                attrs.pop(key)
                setattr(value, '_name', key)
                if isinstance(value, BaseForm) and not isinstance(value, FormMixin):
                    value.__class__ = type(value.__class__.__name__, (FormMixin, value.__class__), {})
                    value.error_class = FormsetErrorList
                attrs['declared_holders'][key] = value

        new_class = super().__new__(cls, name, bases, attrs)

        # Walk through the MRO.
        declared_holders = {}
        for base in reversed(new_class.__mro__):
            # Collect Form and FormCollection instances from base classes.
            if hasattr(base, 'declared_holders'):
                declared_holders.update(base.declared_holders)

            # Form shadowing.
            for attr, value in base.__dict__.items():
                if value is None and attr in declared_holders:
                    declared_holders.pop(attr)

        new_class.declared_holders = declared_holders

        return new_class


class BaseFormCollection(HolderMixin, RenderableMixin):
    """
    The main implementation of all the FormCollection logic.
    """
    default_renderer = None
    prefix = None
    template_name = 'formset/default/collection.html'
    min_siblings = None
    max_siblings = None
    extra_siblings = None
    is_sortable = None
    legend = None
    add_label = None
    ignore_marked_for_removal = None

    def __init__(self, data=None, initial=None, renderer=None, prefix=None, min_siblings=None,
                 max_siblings=None, extra_siblings=None, is_sortable=None):
        self.data = MultiValueDict() if data is None else data
        self.initial = initial
        if prefix is not None:
            self.prefix = prefix
        self._errors = None  # Stores the errors after `clean()` has been called.
        if min_siblings is not None:
            self.min_siblings = min_siblings
        if max_siblings is not None:
            self.max_siblings = max_siblings
        if extra_siblings is not None:
            self.extra_siblings = extra_siblings
        if self.has_many:
            if self.min_siblings is None:
                self.min_siblings = 1
            if self.extra_siblings is None:
                self.extra_siblings = 0
            if is_sortable is not None:
                self.is_sortable = is_sortable
        else:
            self.is_sortable = False

        # Initialize form renderer. Use a global default if not specified
        # either as an argument or as self.default_renderer.
        if renderer is None:
            renderer = self.default_renderer
            if isinstance(self.default_renderer, type):
                renderer = renderer()
        self.renderer = renderer

    def iter_single(self):
        for name, declared_holder in self.declared_holders.items():
            prefix = f'{self.prefix}.{name}' if self.prefix else name
            initial = None
            if isinstance(self.initial, dict):
                initial = self.initial.get(name)
            if initial is None:
                initial = declared_holder.initial
            holder = declared_holder.replicate(
                initial=initial,
                prefix=prefix,
                renderer=self.renderer,
                ignore_marked_for_removal=self.ignore_marked_for_removal,
            )
            holder.is_single = True
            yield holder

    def iter_many(self):
        num_siblings = max(self.min_siblings, self.extra_siblings)
        if self.initial:
            if not isinstance(self.initial, list):
                errmsg = "{class_name} is declared to have siblings, but provided argument `{argument}` is not a list"
                raise FormCollectionError(errmsg.format(class_name=self.__class__.__name__, argument='initial'))
            num_siblings = min(self.max_siblings, max(num_siblings, len(self.initial)) + self.extra_siblings)

        first, last = 0, len(self.declared_holders.items()) - 1
        # add initialized collections/forms
        for position in range(num_siblings):
            for item_num, (name, declared_holder) in enumerate(self.declared_holders.items()):
                prefix = f'{self.prefix}.{position}.{name}' if self.prefix else f'{position}.{name}'
                initial = None
                if self.initial and position < len(self.initial):
                    initial = self.initial[position].get(name)
                if initial is None:
                    initial = declared_holder.initial
                holder = declared_holder.replicate(
                    initial=initial,
                    prefix=prefix,
                    renderer=self.renderer,
                    ignore_marked_for_removal=self.ignore_marked_for_removal,
                )
                holder.position = position
                if item_num == first:
                    holder.is_first = True
                if item_num == last:
                    holder.is_last = True
                yield holder
        # add empty placeholder as template for extra collections
        for item_num, (name, declared_holder) in enumerate(self.declared_holders.items()):
            if self.prefix:
                count = self.prefix.count('${position')
                assert count < 10, "Maximum number of nested FormCollections reached"
                # this context rewriting is necessary to render nested templates properly
                position = f'${{position_{count}}}' if count > 0 else '${position}'
                prefix = f'{self.prefix}.{position}.{name}'
            else:
                position = '${position}'
                prefix = f'${{position}}.{name}'
            holder = declared_holder.replicate(
                prefix=prefix,
                renderer=self.renderer,
                ignore_marked_for_removal=self.ignore_marked_for_removal,
            )
            holder.is_template = True
            holder.position = position
            if item_num == first:
                holder.is_first = True
            if item_num == last:
                holder.is_last = True
            yield holder

    def __iter__(self):
        if self.has_many:
            yield from self.iter_many()
        else:
            yield from self.iter_single()

    def get_context(self):
        return {
            'collection': self,
        }

    @property
    def errors(self):
        """Return an ErrorDict for the data provided for this form collection."""
        if self._errors is None:
            self.full_clean()
        return self._errors

    def is_valid(self):
        """Return True if all forms in this collection are valid."""
        return not self.errors

    def full_clean(self):
        if self.has_many:
            self.cleaned_data = []
            self._errors = ErrorList()
            for index, data in enumerate(self.data):
                cleaned_data = {}
                for name, declared_holder in self.declared_holders.items():
                    if name in data:
                        holder = declared_holder.replicate(
                            data=data[name],
                            ignore_marked_for_removal=self.ignore_marked_for_removal,
                        )
                        if holder.ignore_marked_for_removal and MARKED_FOR_REMOVAL in holder.data:
                            break
                        if holder.is_valid():
                            cleaned_data[name] = holder.cleaned_data
                        else:
                            self._errors.extend([{}] * (index - len(self._errors)))
                            self._errors.append({name: holder.errors})
                    else:
                        # can only happen, if client bypasses browser control
                        self._errors.extend([{}] * (index - len(self._errors)))
                        self._errors.append({name: {NON_FIELD_ERRORS: ["Form data is missing."]}})
                else:
                    self.cleaned_data.append(cleaned_data)
            if len(self.cleaned_data) < self.min_siblings:
                # can only happen, if client bypasses browser control
                self._errors.clear()
                self._errors.append({COLLECTION_ERRORS: ["Too few siblings."]})
            if self.max_siblings and len(self.cleaned_data) > self.max_siblings:
                # can only happen, if client bypasses browser control
                self._errors.clear()
                self._errors.append({COLLECTION_ERRORS: ["Too many siblings."]})
        else:
            self.cleaned_data = {}
            self._errors = ErrorDict()
            for name, declared_holder in self.declared_holders.items():
                if name in self.data:
                    holder = declared_holder.replicate(
                        data=self.data[name],
                        ignore_marked_for_removal=self.ignore_marked_for_removal,
                    )
                    if holder.is_valid():
                        self.cleaned_data[name] = holder.cleaned_data
                    else:
                        self._errors.update({name: holder.errors})
                else:
                    # can only happen, if client bypasses browser control
                    self._errors.update({name: {NON_FIELD_ERRORS: ["Form data is missing."]}})

    def clean(self):
        return self.cleaned_data

    @property
    def has_many(self):
        """
        Returns True if current FormCollection manages a list of sibling forms/(sub-)collections.
        """
        return not (self.min_siblings is None and self.max_siblings is None and self.extra_siblings is None)

    def render(self, template_name=None, context=None, renderer=None):
        if not (renderer or self.renderer):
            renderer = FormRenderer()
        return super().render(template_name, context, renderer)

    def model_to_dict(self, main_object):
        """
        Create initial data from a main object. This then is used to fill the initial data from all its child
        collections and forms.
        Forms which do not correspond to the model given by the main object, are themselves responsible to
        access the proper referenced models.
        """
        object_data = {}
        for name, holder in self.declared_holders.items():
            if callable(getattr(holder, 'model_to_dict', None)):
                object_data[name] = holder.model_to_dict(main_object)
            elif isinstance(holder, BaseModelForm):
                opts = holder._meta
                object_data[name] = model_to_dict(main_object, opts.fields, opts.exclude)
            else:
                object_data[name] = model_to_dict(main_object)
        return object_data

    def construct_instance(self, main_object, cleaned_data):
        """
        Construct the main object and its related objects from the nested dictionary of cleaned data.
        Forms which do not correspond to the model given by the main object, are responsible themselves
        to store the corresponding data inside their related models.
        """
        for name, holder in self.declared_holders.items():
            if callable(getattr(holder, 'construct_instance', None)):
                holder.construct_instance(main_object, self.cleaned_data[name])
            elif isinstance(holder, BaseModelForm):
                opts = holder._meta
                holder.cleaned_data = self.cleaned_data[name]
                holder.instance = main_object
                construct_instance(holder, main_object, opts.fields, opts.exclude)
                holder.save()

    __str__ = render
    __html__ = render


class FormCollection(BaseFormCollection, metaclass=FormCollectionMeta):
    """
    Base class for a collection of forms. Attributes of this class which inherit from
    `django.forms.forms.BaseForm` are managed by this class.
    """
    def get_field(self, path):
        path = path.split('.', 1)
        key, path = path
        return self.holders[key].get_field(path)



from base64 import b16encode
from functools import reduce
from operator import or_
import os
import struct

from django.core.exceptions import ImproperlyConfigured
from django.core.files.uploadedfile import UploadedFile
from django.core.files.storage import default_storage
from django.core.signing import get_cookie_signer
from django.db.models.query_utils import Q
from django.forms.models import ModelChoiceIterator, ModelChoiceIteratorValue
from django.forms.widgets import FileInput, Select, SelectMultiple
from django.utils.translation import gettext_lazy as _
from django.utils.timezone import now, datetime, utc


class IncompleteSelectMixin:
    """
    Extra interfaces for widgets not loading the complete set of choices.
    """

    choices = ()
    max_prefetch_choices = 250
    search_lookup = None

    def __init__(self, attrs=None, choices=(), search_lookup=None):
        if search_lookup:
            self.search_lookup = search_lookup
        if isinstance(self.search_lookup, str):
            self.search_lookup = [self.search_lookup]
        super().__init__(attrs, choices)

    def search(self, search_term):
        try:
            query = reduce(or_, (Q(**{sl: search_term}) for sl in self.search_lookup))
        except TypeError:
            raise ImproperlyConfigured(f"Invalid attribute 'search_lookup' in {self.__class__}.")
        return self.choices.queryset.filter(query)

    def get_context(self, name, value, attrs):
        if isinstance(self.choices, ModelChoiceIterator):
            self.optgroups = self._optgroups_model_choice
            if self.choices.queryset.count() > self.max_prefetch_choices:
                attrs = dict(attrs, incomplete=True)
        else:
            self.optgroups = self._optgroups_static_choice
        context = super().get_context(name, value, attrs)
        options = context['widget'].pop('optgroups')
        context['widget']['options'] = options
        return context

    def _optgroups_static_choice(self, name, values, attrs=None):
        options = []
        for val, label in self.choices:
            val = str(val)
            if val in values:
                options.append({'value': val, 'label': label, 'selected': True})
            else:
                options.append({'value': val, 'label': label})
        return options

    def _optgroups_model_choice(self, name, values, attrs=None):
        values_set = set(values)
        options = [{}] * len(values_set)
        counter = 0
        for val, label in self.choices:
            if not isinstance(val, ModelChoiceIteratorValue):
                continue
            val = str(val)
            if val in values_set:
                index = values.index(val)
                options[index] = {'value': val, 'label': label, 'selected': True}
                values_set.remove(val)
            elif counter < self.max_prefetch_choices:
                options.append({'value': val, 'label': label})
            elif not values_set:
                break
            counter += 1
        return options


class Selectize(IncompleteSelectMixin, Select):
    """
    Render widget suitable for TomSelect
    """
    template_name = 'formset/default/widgets/selectize.html'
    placeholder = _("Select")

    def __init__(self, attrs=None, choices=(), search_lookup=None, placeholder=None):
        super().__init__(attrs, choices, search_lookup)
        if placeholder is not None:
            self.placeholder = placeholder

    def build_attrs(self, base_attrs, extra_attrs):
        attrs = super().build_attrs(base_attrs, extra_attrs)
        if self.is_required:
            attrs['required'] = True  # Selectize overrides the default behaviour
        return attrs

    def _optgroups_static_choice(self, name, values, attrs=None):
        options = [{'value': '', 'label': self.placeholder}]
        options.extend(super()._optgroups_static_choice(name, values, attrs))
        return options

    def _optgroups_model_choice(self, name, values, attrs=None):
        options = [{'value': '', 'label': self.placeholder}]
        options.extend(super()._optgroups_model_choice(name, values, attrs))
        return options


class SelectizeMultiple(Selectize):
    allow_multiple_selected = True
    max_items = 5

    def __init__(self, max_items=None, **kwargs):
        super().__init__(**kwargs)
        if max_items:
            self.max_items = max_items

    def build_attrs(self, base_attrs, extra_attrs):
        attrs = super().build_attrs(base_attrs, extra_attrs)
        attrs['max_items'] = self.max_items
        return attrs


class DualSelector(IncompleteSelectMixin, SelectMultiple):
    """
    Render widget suitable for the <select is="django-dual-selector"> widget
    """
    template_name = 'formset/default/widgets/dual_selector.html'


class DualSortableSelector(DualSelector):
    """
    Render widget suitable for the <select is="django-dual-sortable-selector"> widget
    """
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['is_sortable'] = True
        return context


class UploadedFileInput(FileInput):
    """
    Widget to be used as a replacement for fields of type :class:`django.forms.fields.FileField`
    and :class:`django.forms.fields.ImageField`.
    It converts the submitted POST data to reference the already uploaded file in the directory
    configured for temporary uploads.
    """
    template_name = 'formset/default/widgets/file.html'

    def format_value(self, value):
        return value

    def value_from_datadict(self, data, files, name):
        handle = data.get(name)
        if isinstance(handle, (UploadedFile, bool)):
            return handle
        if hasattr(handle, '__iter__'):
            handle = next(iter(handle), None)
        if isinstance(handle, dict):
            if not handle:
                return False  # marked as deleted
            if 'upload_temp_name' not in handle:
                return
            # file has just been uploaded
            signer = get_cookie_signer(salt='formset')
            upload_temp_name = signer.unsign(handle['upload_temp_name'])
            file = open(default_storage.path(upload_temp_name), 'rb')
            file.seek(0, os.SEEK_END)
            size = file.tell()
            file.seek(0)
            # create pseudo unique prefix to avoid file name collisions
            epoch = datetime(2022, 1, 1, tzinfo=utc)
            prefix = b16encode(struct.pack('f', (now() - epoch).total_seconds())).decode('utf-8')
            filename = '.'.join((prefix, handle['name']))
            files[name] = UploadedFile(
                file=file, name=filename, size=size, content_type=handle['content_type'],
                content_type_extra=handle['content_type_extra'],
            )
        return files.get(name)



import copy
import types

from django.forms.renderers import DjangoTemplates


class FormRenderer(DjangoTemplates):
    max_options_per_line = 0

    _template_mapping = {
        'django/forms/errors/list/default.html': 'formset/default/field_errors.html',
        'django/forms/default.html': 'formset/default/form.html',
        'django/forms/widgets/file.html': 'formset/default/widgets/file.html',
        'django/forms/widgets/radio.html': 'formset/default/widgets/multiple_input.html',
        'django/forms/widgets/checkbox_select.html': 'formset/default/widgets/multiple_input.html',
    }

    def __init__(self, field_css_classes=None, label_css_classes=None, control_css_classes=None,
                 form_css_classes=None, fieldset_css_classes=None, collection_css_classes=None,
                 max_options_per_line=None):
        self.field_css_classes = field_css_classes
        self.label_css_classes = label_css_classes
        self.control_css_classes = control_css_classes
        self.form_css_classes = form_css_classes
        self.fieldset_css_classes = fieldset_css_classes
        self.collection_css_classes = collection_css_classes
        if max_options_per_line is not None:
            self.max_options_per_line = max_options_per_line
        super().__init__()

    def get_template(self, template_name):
        template_name = self._template_mapping.get(template_name, template_name)
        return super().get_template(template_name)

    def _amend_form(self, context):
        context.update(
            control_css_classes=self.control_css_classes,
            form_css_classes=self.form_css_classes,
        )
        return context

    def _amend_label(self, context, hide_checkbox_label=False):
        if self.label_css_classes:
            if not isinstance(context['attrs'], dict):
                context['attrs'] = {}
            context['attrs']['class'] = self.label_css_classes
        if hide_checkbox_label and context['field'].widget_type == 'checkbox':
            # `<label>Label:</label>` is rendered by `{{ field }}`, so remove it to
            # prevent double rendering.
            context.pop('label', None)
            context['attrs'].pop('for', None)
            context['use_tag'] = bool(self.control_css_classes)
        return context

    def _amend_multiple_input(self, context):
        """
        Inlines a small number of radio/checkbox fields to render them on one line.
        """
        max_options = 0
        for group, options, index in context['widget']['optgroups']:
            if group is None:
                max_options = len(context['widget']['optgroups'])
                break
            max_options = max(max_options, len(options))
        context['widget']['inlined_options'] = max_options <= self.max_options_per_line
        return context

    def _amend_fieldset(self, context):
        context.update(
            css_classes=self.fieldset_css_classes,
        )
        return context

    def _amend_collection(self, context):
        context.update({
            'add_collection_button': 'formset/default/buttons/add_collection.html',
            'remove_collection_button': 'formset/default/buttons/remove_collection.html',
            'css_classes': self.collection_css_classes,
            'add_collection_label': context['collection'].add_label,
        })
        return context

    _context_modifiers = {
        'django/forms/default.html': _amend_form,
        'django/forms/label.html': _amend_label,
        'django/forms/widgets/checkbox_select.html': _amend_multiple_input,
        'django/forms/widgets/radio.html': _amend_multiple_input,
        'formset/default/fieldset.html': _amend_fieldset,
        'formset/default/collection.html': _amend_collection,
    }

    @classmethod
    def _copy_context(cls, context):
        """"
        Make a semi-deep copy of context. This is required since the amend-methods
        modify the context before rendering. Python's copy.deepcopy doesn't work here
        because File fields _io.BufferedReader can't be pickled.
        """
        replica = context.copy()
        if 'attrs' in context:
            replica['attrs'] = copy.deepcopy(context['attrs'])
        if 'widget' in context:
            replica['widget'] = copy.deepcopy(context['widget'])
        return replica

    def render(self, template_name, context, request=None):
        context = self._copy_context(context)
        context_modifier = self._context_modifiers.get(template_name)
        if callable(context_modifier):
            context = types.MethodType(context_modifier, self)(context)
        template = self.get_template(template_name)
        return template.render(context, request=request).strip()



#!/usr/bin/env python
from setuptools import setup, find_packages
from formset import __version__


with open('README.md') as fh:
    long_description = fh.read()


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Web Environment',
    'Framework :: Django',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Framework :: Django :: 4.0',
    'Framework :: Django :: 4.1',
]

setup(
    name='django-formset',
    version=__version__,
    description='Prevalidate Django Forms in the browser',
    author='Jacob Rief',
    author_email='jacob.rief@gmail.com',
    url='https://github.com/jrief/django-formset',
    packages=find_packages(exclude=['testapp', 'docs']),
    install_requires=[
        'django>=4.0',
    ],
    extra_requires={
        'thumbnail': ['Pillow'],
    },
    license='MIT',
    platforms=['OS Independent'],
    keywords=['Django Forms', 'webcomponent'],
    classifiers=CLASSIFIERS,
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    zip_safe=False,
)



from django.core import validators
from django.core.exceptions import ImproperlyConfigured
from django.db.models.fields.files import FieldFile
from django.forms import boundfield
from django.forms.fields import FileField
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from formset.widgets import UploadedFileInput
from formset.upload import get_file_info


class CheckboxInputMixin:
    """
    This hack is required for adding the field's label to the rendering context.
    This is to make the single checkbox to be rendered with its label after the input field.
    """
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['label'] = self.label
        return context


class BoundField(boundfield.BoundField):
    @property
    def errors(self):
        errors = self.form.errors.get(self.name, self.form.error_class())
        errors.client_messages = self._get_client_messages()
        return errors

    def as_widget(self, widget=None, attrs=None, only_initial=False):
        widget = widget or self.field.widget
        if self.widget_type == 'checkbox' and not isinstance(widget, CheckboxInputMixin):
            widget.__class__ = type(widget.__class__.__name__, (CheckboxInputMixin, widget.__class__), {'label': self.label})
        if self.field.localize:
            widget.is_localized = True
        attrs = attrs or {}
        attrs = self.build_widget_attrs(attrs, widget)
        if self.auto_id and 'id' not in widget.attrs:
            attrs.setdefault('id', self.html_initial_id if only_initial else self.auto_id)
        return widget.render(
            name=self.name,  # `self.html_name` contains the full path and hence doesn't work with form collections
            value=self.value(),
            attrs=attrs,
            renderer=self.form.renderer,
        )

    def build_widget_attrs(self, attrs, widget=None):
        attrs = super().build_widget_attrs(attrs, widget)
        if self.form.form_id:
            attrs['form'] = self.form.form_id
        if hasattr(self.field, 'regex'):
            attrs['pattern'] = self.field.regex.pattern
        return attrs

    def css_classes(self, extra_classes=None):
        """
        Return a string of space-separated CSS classes for this field.
        """
        if hasattr(extra_classes, 'split'):
            extra_classes = extra_classes.split()
        extra_classes = set(extra_classes or [])
        if self.field.required:
            if self.widget_type == 'checkboxselectmultiple':
                extra_classes.add('dj-required-any')
            else:
                extra_classes.add('dj-required')

        # field_css_classes is an optional member of a FormRenderer optimized for django-formset
        field_css_classes = getattr(self.form.renderer, 'field_css_classes', None)
        if isinstance(field_css_classes, dict):
            try:
                field_css_classes = field_css_classes[self.name]
            except KeyError:
                field_css_classes = field_css_classes.get('*')
        if hasattr(field_css_classes, 'split'):
            extra_classes.update(field_css_classes.split())
        elif isinstance(field_css_classes, (list, tuple)):
            extra_classes.update(field_css_classes)
        return super().css_classes(extra_classes)

    @cached_property
    def widget_type(self):
        return super().widget_type

    @cached_property
    def auto_id(self):
        """
        Since we can have many forms with a different name each, prefix the id with the form name
        """
        auto_id = self.form.auto_id
        if auto_id and '%s' in str(auto_id):
            auto_id = auto_id % self.html_name
            if getattr(self.form, 'name', None):
                return f'{self.form.name}_{auto_id}'
            return auto_id
        elif auto_id:
            return self.html_name
        return ''

    def value(self):
        value = super().value()
        if isinstance(value, FieldFile):
            return get_file_info(value)
        return value

    def _get_client_messages(self):
        """
        Extract server validation error messages to be rendered by the client side.
        """
        client_messages = {}
        server_messages = self.field.error_messages
        if self.field.required is True:
            if self.widget_type == 'checkboxselectmultiple':
                client_messages['custom_error'] = _("At least one checkbox must be selected.")
            elif 'required' in server_messages:
                if self.widget_type == 'selectize':
                    client_messages['custom_error'] = server_messages['required']
                else:
                    client_messages['value_missing'] = server_messages['required']
        if 'invalid' in server_messages:
            client_messages['type_mismatch'] = client_messages['pattern_mismatch'] = server_messages['invalid']
        elif 'invalid_choice' in server_messages:
            client_messages['type_mismatch'] = server_messages['invalid_choice']
        else:
            for validator in self.field.validators:
                validator_code = getattr(validator, 'code', None)
                if validator_code == 'invalid':
                    client_messages['type_mismatch'] = validator.message
        if getattr(self.field, 'max_length', None) is not None:
            data = {'max_length': self.field.max_length}
            max_length_message = _("Ensure this value has at most %(max_length)s characters.")
            if isinstance(self.field, FileField):
                client_messages['too_long'] = max_length_message % data
            else:
                client_messages['too_long'] = server_messages.get('max_length', max_length_message) % data
        if getattr(self.field, 'min_length', None) is not None:
            data = {'min_length': self.field.min_length}
            min_length_message = _("Ensure this value has at least %(min_length)s characters.")
            client_messages['too_short'] = server_messages.get('min_length', min_length_message) % data
        if getattr(self.field, 'min_value', None) is not None:
            data = {'limit_value': self.field.min_value}
            client_messages['range_underflow'] = server_messages.get('min_value', validators.MinValueValidator.message) % data
        if getattr(self.field, 'max_value', None) is not None:
            data = {'limit_value': self.field.max_value}
            client_messages['range_overflow'] = server_messages.get('max_value', validators.MaxValueValidator.message) % data
        try:
            step_value = float(self.field.widget.attrs['step'])
        except (KeyError, TypeError):
            pass
        else:
            data = {'step_value': step_value}
            client_messages['step_mismatch'] = _("Input value must be a multiple of {step_value}.").format(**data)
        client_messages['bad_input'] = validators.ProhibitNullCharactersValidator.message
        if isinstance(self.field, FileField):
            if not isinstance(self.field.widget, UploadedFileInput):
                raise ImproperlyConfigured(
                    f"Field of type {self.field.__class__} must use widget inheriting from {UploadedFileInput}"
                )
            # abuse built-in client errors for failed file upload messages
            client_messages.update(
                type_mismatch=_("File upload still in progress."),
                bad_input=_("File upload failed."),
            )
        return client_messages
